@@ -1,4 +1,5 @@
 from quickstats._version import __version__
+from quickstats.decorators import *
 
 import pathlib
 
@@ -9,4 +10,9 @@ macro_path = pathlib.Path(__file__).parent.absolute()
 
 MAX_WORKERS = 8
 
-from quickstats.main import *
\ No newline at end of file
+from quickstats.utils.io import VerbosePrint
+
+_PRINT_ = VerbosePrint("INFO")
+
+from quickstats.main import *
+
