@@ -196,6 +196,7 @@ def likelihood_scan(**kwargs):
                                                                    'instead of using the probability '
                                                                    'density at the bin centre')
 @click.option('--constrain/--no-constrain', 'constrain_nuis', default=True, help='Use constrained NLL')
+@click.option('-v', '--verbosity',  default="INFO", help='verbosity level ("DEBUG", "INFO", "WARNING", "ERROR")')
 def cls_limit(**kwargs):
     """
     Tool for evaluating Asymptotic CLs limit
@@ -294,6 +295,8 @@ def generate_asimov(**kwargs):
 @click.option('-o', '--output_file', 'outname', required=True, 
               help='Name of the output workspace containing the '
                    'generated asimov dataset.')
+@click.option('-d', '--data', 'data_name', default='combData', show_default=True,
+              help='Name of the dataset used in NP profiling.')
 @click.option('-p', '--poi', 'poi_name', required=True, 
               help='Name of the parameter of interest (POI).')
 @click.option('-s', '--poi_scale', type=float, default=1.0, show_default=True,
@@ -318,4 +321,49 @@ def generate_standard_asimov(**kwargs):
     asimov_types = [int(t) for t in asimov_types.split(',')]
     generator = AsimovGenerator(**kwargs)
     generator.generate(asimov_types)
-    generator.save(outname)
\ No newline at end of file
+    generator.save(outname)
+    
+
+@click.command(name='toy_significance')
+@click.option('-i', '--input_file', 'filename', required=True, 
+              help='Path to the input workspace file.')
+@click.option('-o', '--output_file', 'outname', default="toy_study/results.json", 
+              help='Name of the output file containing toy results.')
+@click.option('-n', '--n_toys', type=int,
+              help='Number of the toys to use.')
+@click.option('-b', '--batchsize', type=int, default=100, show_default=True,
+              help='Divide the task into batches each containing this number of toys. '
+                   'Result from each batch is saved for caching and different batches '
+                   'are run in parallel if needed')
+@click.option('-s', '--seed', type=int, default=0,  show_default=True,
+              help='Random seed used for generating toy datasets.')
+@click.option('-p', '--poi', 'poi_name', default=None,
+              help='Name of the parameter of interest (POI). If None, the first POI is used.')
+@click.option('-v', '--poi_val', type=float, default=0,  show_default=True,
+              help='POI value when generating the toy dataset.')
+@click.option('--binned/--unbinned', default=True, show_default=True,
+              help='Generate binned toy dataset.')
+@click.option('--cache/--no-cache', default=True,  show_default=True,
+              help='Cache existing batch results.')
+@click.option('--fit_options', default=None, help='A json file specifying the fit options.')
+@click.option('--verbosity', default="ERROR", help='Quickstats internal verbosity level.')
+@click.option('--parallel', type=int, default=-1, help='Parallelize job across different scan values.'+\
+                                                       'Use -1 for N_CPU workers.')
+def toy_significance(**kwargs):
+    """
+    Generate toys and evaluate significance
+    """
+    from quickstats.components import PValueToys
+    n_toys = kwargs.pop("n_toys")
+    batchsize = kwargs.pop("batchsize")
+    seed = kwargs.pop("seed")
+    cache = kwargs.pop("cache")
+    outname = kwargs.pop("outname")
+    parallel = kwargs.pop("parallel")
+    pvalue_toys = PValueToys(**kwargs)
+    pvalue_toys.get_toy_results(n_toys=n_toys, batchsize=batchsize, seed=seed,
+                                cache=cache, save_as=outname, parallel=parallel)
+    
+    
+    
+    
