@@ -1,3 +1,4 @@
+from quickstats.components.abstract_object import AbstractObject
 from quickstats.components.extended_minimizer import ExtendedMinimizer
 from quickstats.components.extended_model import ExtendedModel
 from quickstats.components.analysis_object import AnalysisObject
@@ -7,4 +8,6 @@ from quickstats.components.nuisance_parameter_pull import NuisanceParameterPull
 from quickstats.components.roofit_plotter import RooFitPlotter
 from quickstats.components.nuisance_parameter_harmonizer import NuisanceParameterHarmonizer
 from quickstats.components.asymptotic_cls import AsymptoticCLs
-from quickstats.components.likelihood import Likelihood
\ No newline at end of file
+from quickstats.components.likelihood import Likelihood
+from quickstats.components.pvalue_toys import PValueToys
+from quickstats.components.toy_limit_calculator import ToyLimitCalculator
\ No newline at end of file
