@@ -1,25 +1,28 @@
-from typing import Optional, Union
+from typing import Optional, Union, List
 
+from quickstats import _PRINT_
 from quickstats.components import ExtendedModel, ExtendedMinimizer
 import ROOT
 
 class AnalysisObject(object):
     def __init__(self, filename:Optional[str]=None, data_name:str='combData', binned_likelihood:bool=True,
                  fix_param:str='', profile_param:str='', ws_name:Optional[str]=None, 
-                 mc_name:Optional[str]=None, snapshot_name:Optional[str]=None,
+                 mc_name:Optional[str]=None, snapshot_name:Optional[Union[List[str], str]]=None,
                  minimizer_type:str='Minuit2', minimizer_algo:str='Migrad', precision:float=0.001, 
                  eps:float=1.0, strategy:int=1, print_level:int=-1, timer:bool=False,
                  num_cpu:int=1, offset:bool=True, optimize:int=2, eigen:bool=False,
                  fix_cache:bool=True, fix_multi:bool=True, max_calls:int=-1, 
                  max_iters:int=-1, constrain_nuis:bool=True, batch_mode:bool=False,
                  int_bin_precision:float=-1., verbosity:Union[int, str]="INFO", **kwargs):
-        
+
         self.model = None
         self.minimizer = None
         
+        _PRINT_.verbosity = verbosity
+        
         if filename is not None:
             self.setup_model(filename, data_name, binned_likelihood, ws_name, mc_name, 
-                             snapshot_name, fix_cache, fix_multi, verbosity=verbosity)
+                             snapshot_name, fix_cache, fix_multi)
             self.setup_parameters(fix_param, profile_param)
             self.setup_minimizer(minimizer_type, minimizer_algo, precision, eps, strategy, num_cpu,
                                  offset, optimize, eigen, max_calls, max_iters, print_level,
@@ -28,12 +31,11 @@ class AnalysisObject(object):
     def setup_model(self, filename, data_name='combData', binned_likelihood:bool=True, 
                     ws_name:Optional[str]=None, mc_name:Optional[str]=None, 
                     snapshot_name:Optional[str]=None, fix_cache:bool=True, 
-                    fix_multi:bool=True, verbosity:Union[int, str]="INFO"):
+                    fix_multi:bool=True):
         
         model = ExtendedModel(fname=filename, ws_name=ws_name, mc_name=mc_name, data_name=data_name, 
                               snapshot_name=snapshot_name, binned_likelihood=binned_likelihood,
-                              fix_cache=fix_cache, fix_multi=fix_multi,
-                              verbosity=verbosity)
+                              fix_cache=fix_cache, fix_multi=fix_multi)
         self.model = model
     
     def setup_parameters(self, fix_param:str='', profile_param:str=''):
@@ -50,7 +52,7 @@ class AnalysisObject(object):
                         max_calls:int=-1, max_iters:int=-1, print_level:int=-1, timer:bool=False,
                         constrain_nuis:bool=True, batch_mode:bool=False, int_bin_precision:float=-1.):
         
-        minimizer = ExtendedMinimizer("minimizer", self.model.pdf, self.model.data)
+        minimizer = ExtendedMinimizer("Minimizer", self.model.pdf, self.model.data)
         
         nll_commands = [ROOT.RooFit.NumCPU(num_cpu, 3), 
                         ROOT.RooFit.Offset(offset)]
