@@ -35,13 +35,14 @@ class AsymptoticCLs(AnalysisObject):
                  num_cpu:int=1, offset:bool=True, optimize:int=2,
                  fix_cache:bool=True, fix_multi:bool=True, max_calls:int=-1,
                  max_iters:int=-1, constrain_nuis:bool=True, batch_mode:bool=False,
-                 int_bin_precision:float=-1., **kwargs):
+                 int_bin_precision:float=-1., verbosity:Union[int, str]="INFO", **kwargs):
         """
         args:
             precision: precision in mu that defines iterative cutoff
             do_tilde: bound mu at zero if true and do the \tilde{q}_{mu} asymptotics
             use_predictive_fit: experimental, extrapolate best fit nuisance parameters based on previous fit results
         """
+
         config = {
             'filename': filename,
             'data_name': data_name,
@@ -67,7 +68,8 @@ class AsymptoticCLs(AnalysisObject):
             'max_iters': max_iters,
             'constrain_nuis': constrain_nuis,
             'batch_mode': batch_mode,
-            'int_bin_precision': int_bin_precision
+            'int_bin_precision': int_bin_precision,
+            'verbosity': verbosity
         }
         
         super().__init__(**config)
@@ -448,7 +450,8 @@ class AsymptoticCLs(AnalysisObject):
         snapshot_0_name = self.nll_maps[asimov_0_nll]['snapshot']
         self.load_snapshot(snapshot_0_name)
         asimov_data_n = self.model.generate_asimov(self.poi.GetName(), poi_val=n_times_sigma,
-                                                   do_fit=False, modify_globs=True)
+                                                   do_fit=False, modify_globs=True,
+                                                   restore_states=1)
         asimov_n_nll = self.create_nll(asimov_data_n)
         self.load_conditional_snapshot(n_times_sigma, target="globs")
         self.load_conditional_snapshot(n_times_sigma, target="nuis")
@@ -482,7 +485,8 @@ class AsymptoticCLs(AnalysisObject):
                                                    poi_profile=0, 
                                                    do_fit=not self.do_blind,
                                                    modify_globs=not self.do_blind,
-                                                   dataset=self.data)
+                                                   dataset=self.data,
+                                                   restore_states=1)
         asimov_0_nll = self.create_nll(asimov_data_0)
         self.summary[0] = {'dataset': asimov_data_0, 'nll': asimov_0_nll}
         self.set_poi_value(0)
