@@ -15,13 +15,12 @@ import numpy as np
 import ROOT
 
 import quickstats
+from quickstats import semistaticmethod
+from quickstats.components import AbstractObject
 from quickstats.components.numerics import is_integer, pretty_value
 from quickstats.utils.root_utils import load_macro
-from quickstats.utils.io import VerbosePrint
 
-_PRINT_ = VerbosePrint("INFO")
-
-class ExtendedModel(object):
+class ExtendedModel(AbstractObject):
     
     _DEFAULT_NAMES_ = {
         'conditional_globs': 'conditionalGlobs_{mu}',
@@ -37,23 +36,24 @@ class ExtendedModel(object):
     }
     
     def __init__(self, fname:str, ws_name:Optional[str]=None, mc_name:Optional[str]=None,
-                 data_name:str="combData", snapshot_name:Optional[str]=None,
+                 data_name:Optional[str]="combData", snapshot_name:Optional[Union[List[str], str]]=None,
                  binned_likelihood:bool=True, tag_as_measurement:Optional[str]=None,
                  fix_cache:bool=True, fix_multi:bool=True, interpolation_code:int=-1,
-                 load_extension:bool=True, verbosity:Union[int, str]="INFO"):
+                 load_extension:bool=True, verbosity:Optional[Union[int, str]]=None):
+        super().__init__(verbosity=verbosity)
         self.fname = fname
         self.ws_name = ws_name
         self.mc_name = mc_name
         self.data_name = data_name
-        self.snapshot_name = snapshot_name
+        self.initial_snapshots = snapshot_name
         self.binned_likelihood = binned_likelihood
         self.tag_as_measurement = tag_as_measurement
         self.fix_cache = fix_cache
         self.fix_multi = fix_multi
         self.interpolation_code = interpolation_code
-        _PRINT_.verbosity = verbosity
         if load_extension:
             self.load_extension()
+        
         self.initialize()
   
     @property
@@ -82,15 +82,30 @@ class ExtendedModel(object):
         return self._pois
     @property
     def observables(self):
-        return self._observables        
+        return self._observables
     
-    @staticmethod
-    def load_extension():
+    @property
+    def initial_snapshots(self):
+        return self._initial_snapshots
+    
+    @initial_snapshots.setter
+    def initial_snapshots(self, val):
+        if val is None:
+            self._initial_snapshots = []
+        elif isinstance(val, str):
+            self._initial_snapshots = [s.strip() for s in val.split(',') if s]
+        elif isinstance(val, list):
+            self._initial_snapshots = val
+        else:
+            raise ValueError("`initial_snapshots` must be string or list of strings")
+    
+    @semistaticmethod
+    def load_extension(self):
         try:
             if not hasattr(ROOT, 'RooTwoSidedCBShape'):
                 result = load_macro('RooTwoSidedCBShape')
                 if hasattr(ROOT, 'RooTwoSidedCBShape'):
-                    _PRINT_.info('INFO: Loaded extension module "RooTwoSidedCBShape"')
+                    self.stdout.info('INFO: Loaded extension module "RooTwoSidedCBShape"')
         except Exception as e:
             print(e)
             
@@ -98,8 +113,8 @@ class ExtendedModel(object):
     def load_RooFitObjects():
         load_macro('RooFitObjects')
         
-    @staticmethod
-    def modify_interp_codes(ws, interp_code, classes=None):
+    @semistaticmethod
+    def modify_interp_codes(self, ws, interp_code, classes=None):
         if classes is None:
             classes = [ROOT.RooStats.HistFactory.FlexibleInterpVar, ROOT.PiecewiseInterpolation]
         for component in ws.components():
@@ -107,7 +122,7 @@ class ExtendedModel(object):
                 if (component.IsA() == cls.Class()):
                     component.setAllInterpCodes(interp_code)
                     class_name = cls.Class_Name().split('::')[-1]
-                    _PRINT_.info('INFO: {} {} interpolation code set to {}'.format(component.GetName(),
+                    self.stdout.info('INFO: {} {} interpolation code set to {}'.format(component.GetName(),
                                                                             class_name,
                                                                             interp_code))
         return None
@@ -119,9 +134,8 @@ class ExtendedModel(object):
         if numeric_integration:
             ROOT.RooMsgService.instance().getStream(1).removeTopic(ROOT.RooFit.NumIntegration)
 
-    @staticmethod
-    def activate_binned_likelihood(ws):
-        _PRINT_.info('INFO: Activating binned likelihood evaluation')
+    @semistaticmethod
+    def activate_binned_likelihood(self, ws):
         for component in ws.components():
             try:
                 flag = (component.IsA() == ROOT.RooRealSumPdf.Class())
@@ -129,12 +143,11 @@ class ExtendedModel(object):
                 flag = (component.ClassName() == "RooRealSumPdf")
             if (flag):
                 component.setAttribute('BinnedLikelihood')
-                _PRINT_.info('INFO: Activated binned likelihood attribute for {}'.format(component.GetName()))
+                self.stdout.info('INFO: Activated binned likelihood attribute for {}'.format(component.GetName()))
         return None
                           
-    @staticmethod
-    def set_measurement(ws, condition):
-        _PRINT_.info('INFO: Activating measurements to reduce memory consumption')
+    @semistaticmethod
+    def set_measurement(self, ws, condition):
         for component in ws.components():
             name = component.GetName()
             try:
@@ -143,22 +156,30 @@ class ExtendedModel(object):
                 flag = (component.ClassName() == "RooAddPdf")
             if flag and condition(name):
                 component.setAttribute('MAIN_MEASUREMENT')
-                _PRINT_.info('INFO: Activated main measurement attribute for {}'.format(name))
+                self.stdout.info('INFO: Activated main measurement attribute for {}'.format(name))
         return None
     
-    @staticmethod
-    def deactivate_lv2_const_optimization(ws, condition):
-        _PRINT_.info('INFO: Deactivating level 2 constant term optimization for specified pdfs')
+    @semistaticmethod
+    def deactivate_lv2_const_optimization(self, ws, condition):
+        self.stdout.info('INFO: Deactivating level 2 constant term optimization for specified pdfs')
         for component in ws.components():
             name = component.GetName()
             if (component.InheritsFrom(ROOT.RooAbsPdf.Class()) and condition(name)):
                 component.setAttribute("NOCacheAndTrack")
-                _PRINT_.info('INFO: Deactivated level 2 constant term optimization for {}'.format(name))
+                self.stdout.info('INFO: Deactivated level 2 constant term optimization for {}'.format(name))
+                
+    def load_snapshots(self, snapshot_names:List[str]):
+        for snapshot_name in snapshot_names:
+            if self.workspace.getSnapshot(snapshot_name):
+                self.workspace.loadSnapshot(snapshot_name)
+                self.stdout.info(f'INFO: Loaded snapshot "{snapshot_name}"')
+            else:
+                self.stdout.warning(f'WARNING: Failed to load snapshot "{snapshot_name}"')
             
     def initialize(self):
         if not os.path.exists(self.fname):
             raise FileNotFoundError('workspace file {} does not exist'.format(self.fname))
-        _PRINT_.info('INFO: Opening file "{}"'.format(self.fname))
+        self.stdout.info('INFO: Opening file "{}"'.format(self.fname))
         file = ROOT.TFile(self.fname) 
         if (not file):
             raise RuntimeError("Something went wrong while loading the root file: {}".format(self.fname))
@@ -168,28 +189,28 @@ class ExtendedModel(object):
             if not ws_names:
                 raise RuntimeError("No workspaces found in the root file: {}".format(self.fname))
             if len(ws_names) > 1:
-                _PRINT_.warning("WARNING: Found multiple workspace instances from the root file: {}. Available workspaces"
+                self.stdout.warning("WARNING: Found multiple workspace instances from the root file: {}. Available workspaces"
                       " are \"{}\". Will choose the first one by default".format(self.fname, ','.join(ws_names)))
             self.ws_name = ws_names[0]
 
         ws = file.Get(self.ws_name)
         if not ws:
             raise RuntimeError('failed to load workspace "{}"'.format(self.ws_name))
-        _PRINT_.info('INFO: Loaded workspace "{}"'.format(self.ws_name))
+        self.stdout.info('INFO: Loaded workspace "{}"'.format(self.ws_name))
         # load model config
         if self.mc_name is None:
             mc_names = [i.GetName() for i in ws.allGenericObjects() if 'ModelConfig' in i.ClassName()]
             if not mc_names:
                 raise RuntimeError("no ModelConfig object found in the workspace: {}".format(ws_name))
             if len(mc_names) > 1:
-                _PRINT_.warning("WARNING: Found multiple ModelConfig instances from the workspace: {}. "
+                self.stdout.warning("WARNING: Found multiple ModelConfig instances from the workspace: {}. "
                       "Available ModelConfigs are \"{}\". "
                       "Will choose the first one by default".format(ws_name, ','.join(mc_names)))
             self.mc_name = mc_names[0]
         model_config = ws.obj(self.mc_name)
         if not model_config:
             raise RuntimeError('failed to load model config "{}"'.format(self.mc_name))
-        _PRINT_.info('INFO: Loaded model config "{}"'.format(self.mc_name))
+        self.stdout.info('INFO: Loaded model config "{}"'.format(self.mc_name))
             
         # modify interpolation code
         if self.interpolation_code != -1:
@@ -212,41 +233,46 @@ class ExtendedModel(object):
         pdf = model_config.GetPdf()
         if not pdf:
             raise RuntimeError('Failed to load pdf')
-        _PRINT_.info('INFO: Loaded model pdf "{}" from model config'.format(pdf.GetName()))
+        self.stdout.info('INFO: Loaded model pdf "{}" from model config'.format(pdf.GetName()))
              
         # load dataset
+        if self.data_name is None:
+            data_names = [i.GetName() for i in ws.allData()]
+            if not data_names:
+                raise RuntimeError("no datasets found in the workspace: {}".format(ws_name))
+            self.data_name = data_names[0]
         data = ws.data(self.data_name)
         if not data:
             raise RuntimeError('Failed to load dataset')
-        _PRINT_.info('INFO: Loaded dataset "{}" from workspace'.format(data.GetName()))
+        self.stdout.info('INFO: Loaded dataset "{}" from workspace'.format(data.GetName()))
                 
         # load nuisance parameters
         nuisance_parameters = model_config.GetNuisanceParameters()
         if not nuisance_parameters:
             #raise RuntimeError('Failed to load nuisance parameters')
-            _PRINT_.info("WARNING: No nuisance parameters found in the workspace. "
+            self.stdout.info("WARNING: No nuisance parameters found in the workspace. "
                          "An empty set will be loaded by default.")
             nuisance_parameters = ROOT.RooArgSet()
         else:
-            _PRINT_.info('INFO: Loaded nuisance parameters from model config')
+            self.stdout.info('INFO: Loaded nuisance parameters from model config')
                 
         # Load global observables
         global_observables = model_config.GetGlobalObservables()
         if not global_observables:
             raise RuntimeError('Failed to load global observables')          
-        _PRINT_.info('INFO: Loaded global observables from model config')                  
+        self.stdout.info('INFO: Loaded global observables from model config')                  
     
         # Load POIs
         pois = model_config.GetParametersOfInterest()
         if not pois:
             raise RuntimeError('Failed to load parameters of interest')
-        _PRINT_.info('INFO: Loaded parameters of interest from model config')
+        self.stdout.info('INFO: Loaded parameters of interest from model config')
                                   
         # Load observables
         observables = model_config.GetObservables()
         if not observables:
             raise RuntimeError('Failed to load observables')     
-        _PRINT_.info('INFO: Loaded observables from model config')
+        self.stdout.info('INFO: Loaded observables from model config')
         
         self._file                = file
         self._workspace           = ws
@@ -259,9 +285,7 @@ class ExtendedModel(object):
         self._observables         = observables
                           
         # Load snapshots
-        if (self.snapshot_name) and self.workspace.getSnapshot(self.snapshot_name):
-            self._workspace.loadSnapshot(self.snapshot_name)
-            _PRINT_.info('INFO: Loaded snapshot "{}"'.format(self.snapshot_name))
+        self.load_snapshots(self.initial_snapshots)
         
         self.remove_msg_topic()
         return None
@@ -292,7 +316,7 @@ class ExtendedModel(object):
         profiled_parameters = self._profile_parameters(self.workspace.allVars(), param_expr=param_expr,
                                                        param_str='parameter') 
         if not profiled_parameters:
-            _PRINT_.info('Info: No parameters are profiled.')
+            self.stdout.info('Info: No parameters are profiled.')
         return profiled_parameters 
     
     def fix_nuisance_parameters(self, param_expr=None):
@@ -305,22 +329,22 @@ class ExtendedModel(object):
     def profile_parameters_of_interest(self, param_expr=None):
         return self._profile_parameters(self.pois, param_expr=param_expr, param_str='parameter of interest')
     
-    @staticmethod
-    def _set_parameters(source:"ROOT.RooArgSet", param_dict, mode=None, param_str='parameter'):
+    @semistaticmethod
+    def _set_parameters(self, source:"ROOT.RooArgSet", param_dict, mode=None, param_str='parameter'):
         set_parameters = []
         available_parameters = [param.GetName() for param in source]
         for name in param_dict:
             selected_params = [param for param in available_parameters if fnmatch.fnmatch(param, name)]
             if not selected_params:
-                _PRINT_.warning('WARNING: Parameter "{}" does not exist. No modification will be made.'.format(name))
+                self.stdout.warning('WARNING: Parameter "{}" does not exist. No modification will be made.'.format(name))
             for param_name in selected_params:
                 ExtendedModel._set_parameter(source[param_name], param_dict[name], mode=mode, param_str=param_str)
                 set_parameters.append(source[param_name])
 
         return set_parameters
     
-    @staticmethod
-    def _set_parameter(param, value, mode=None, param_str='parameter'):
+    @semistaticmethod
+    def _set_parameter(self, param, value, mode=None, param_str='parameter'):
         name = param.GetName()
         old_value = param.getVal()
         new_value = old_value
@@ -339,7 +363,7 @@ class ExtendedModel(object):
                 if (new_value < v_min) or (new_value > v_max):
                     new_value = (v_min + v_max)/2
                 param.setRange(v_min, v_max)
-                _PRINT_.info('INFO: Set {} "{}" range to ({},{})'.format(param_str, name, v_min, v_max))
+                self.stdout.info('INFO: Set {} "{}" range to ({},{})'.format(param_str, name, v_min, v_max))
             elif (v_min is not None):
                 if (new_value < v_min):
                     new_value = v_min
@@ -347,7 +371,7 @@ class ExtendedModel(object):
                 if (v_min == 0) and (old_value < 0):
                     new_value = abs(old_value)
                 param.setMin(v_min)
-                _PRINT_.info('INFO: Set {} "{}" min value to ({},{})'.format(param_str, name, v_min))
+                self.stdout.info('INFO: Set {} "{}" min value to ({},{})'.format(param_str, name, v_min))
             elif (v_max is not None):
                 if (new_value > v_max):
                     new_value = v_max
@@ -355,16 +379,16 @@ class ExtendedModel(object):
                 if (v_max == 0) and (old_value > 0):
                     new_value = -abs(old_value)                    
                 param.setMax(v_max)
-                _PRINT_.info('INFO: Set {} "{}" max value to ({},{})'.format(param_str, name, v_max))
+                self.stdout.info('INFO: Set {} "{}" max value to ({},{})'.format(param_str, name, v_max))
         if new_value != old_value:
             param.setVal(new_value)              
-            _PRINT_.info('INFO: Set {} "{}" value to {}'.format(param_str, name, new_value))
+            self.stdout.info('INFO: Set {} "{}" value to {}'.format(param_str, name, new_value))
         if mode=='fix':
             param.setConstant(1)
-            _PRINT_.info('INFO: Fixed {} "{}" at value {}'.format(param_str, name, param.getVal()))
+            self.stdout.info('INFO: Fixed {} "{}" at value {}'.format(param_str, name, param.getVal()))
         elif mode=='free':
             param.setConstant(0)
-            _PRINT_.info('INFO: "{}" = [{}, {}]'.format(name, param.getMin(), param.getMax()))
+            self.stdout.info('INFO: "{}" = [{}, {}]'.format(name, param.getMin(), param.getMax()))
         return None
 
     @staticmethod
@@ -432,6 +456,17 @@ class ExtendedModel(object):
                 raise ValueError('invalid parameter expression: {}'.format(param))
         return param_dict
     
+    @staticmethod
+    def randomize_globs(pdf:ROOT.RooAbsPdf, globs:ROOT.RooDataSet, seed:int):
+        """Randomize values of global observables (for generating pseudo-experiments)
+        """
+        # set random seed for reproducible result
+        if seed >= 0:
+            ROOT.RooRandom.randomGenerator().SetSeed(seed)
+        pseudo_globs = pdf.generateSimGlobal(globs, 1)
+        pdf_vars = pdf.getVariables()
+        pdf_vars.assignValueOnly(pseudo_globs.get(0))
+    
     @staticmethod
     def find_unique_prod_components(root_pdf, components, recursion_count=0):
         if (recursion_count > 50):
@@ -485,8 +520,9 @@ class ExtendedModel(object):
                 result.add(unfolded_components)
         return result
     
-    @staticmethod
-    def _pair_constraints(constraint_set:ROOT.RooArgSet, globs:ROOT.RooArgSet, nuis:ROOT.RooArgSet):
+    @semistaticmethod
+    def _pair_constraints(self, constraint_set:ROOT.RooArgSet, globs:ROOT.RooArgSet, nuis:ROOT.RooArgSet,
+                          base_component:bool=False):
         nuis_list = []
         glob_list = []
         pdf_list = []
@@ -495,47 +531,49 @@ class ExtendedModel(object):
         for pdf in constraint_set:
             target_np   = None
             target_glob = None
-            # old method (not getting correct nuis)
-            """
-            components = pdf.getComponents()
-            components.remove(pdf)
-            if components.getSize():
-                for c1 in components:
-                    for c2 in components:
-                        if c1 == c2:
-                            continue
-                        if c2.dependsOn(c1):
-                            components.remove(c1)
-                if (components.getSize() > 1):
-                    raise RuntimeError("failed to isolate proper nuisance parameter")
-                elif (components.getSize() == 1):
-                    target_np = components.first()
-            else:
-                for np in nuis:
-                    if pdf.dependsOn(np):
-                        target_np = np
-                        break
-            """
-            unfolded_nuis = ExtendedModel.unfold_components(pdf, set(nuis))
-            if not unfolded_nuis:
-                for np in nuis:
-                    if pdf.dependsOn(np):
-                        target_np = np
-                        break
-            elif len(unfolded_nuis) > 1:
-                raise RuntimeError("failed to isolate proper nuisance parameter")
-            else:
-                target_np = unfolded_nuis.first()
             
+            if base_component:
+                # getting base pdf component
+                components = pdf.getComponents()
+                components.remove(pdf)
+                if components.getSize():
+                    for c1 in components:
+                        for c2 in components:
+                            if c1 == c2:
+                                continue
+                            if c2.dependsOn(c1):
+                                components.remove(c1)
+                    if (components.getSize() > 1):
+                        raise RuntimeError("failed to isolate proper nuisance parameter")
+                    elif (components.getSize() == 1):
+                        target_np = components.first()
+                else:
+                    for np in nuis:
+                        if pdf.dependsOn(np):
+                            target_np = np
+                            break
+            else:
+                # getting the actual nuisance parameter
+                unfolded_nuis = ExtendedModel.unfold_components(pdf, set(nuis))
+                if not unfolded_nuis:
+                    for np in nuis:
+                        if pdf.dependsOn(np):
+                            target_np = np
+                            break
+                elif len(unfolded_nuis) > 1:
+                    raise RuntimeError("failed to isolate proper nuisance parameter")
+                else:
+                    target_np = unfolded_nuis.first()
+                    
             if not target_np:
-                _PRINT_.warning('WARNING: Could not find nuisance parameter for the constraint: {}'.format(pdf.GetName()))
+                self.stdout.warning('WARNING: Could not find nuisance parameter for the constraint: {}'.format(pdf.GetName()))
                 continue
             for glob in globs:
                 if pdf.dependsOn(glob):
                     target_glob = glob
                     break
             if not target_glob:
-                _PRINT_.warning('WARNING: Could not find global observable for the constraint: {}'.format(pdf.GetName()))
+                self.stdout.warning('WARNING: Could not find global observable for the constraint: {}'.format(pdf.GetName()))
                 continue
             nuis_list.append(target_np)
             glob_list.append(target_glob)
@@ -547,9 +585,11 @@ class ExtendedModel(object):
         nuis_list, glob_list, _ = self._pair_constraints(constraint_set, self.global_observables, self.nuisance_parameters)
         return nuis_list, glob_list
 
-    def pair_constraints(self, to_str=False):
+    def pair_constraints(self, to_str=False, base_component:bool=False):
         constraint_set = self.unfold_constraints()
-        nuis_list, glob_list, pdf_list = self._pair_constraints(constraint_set, self.global_observables, self.nuisance_parameters)
+        nuis_list, glob_list, pdf_list = self._pair_constraints(constraint_set, self.global_observables, 
+                                                                self.nuisance_parameters,
+                                                                base_component=base_component)
         if to_str:
             nuis_names = [i.GetName() for i in nuis_list]
             glob_names = [i.GetName() for i in glob_list]
@@ -608,7 +648,7 @@ class ExtendedModel(object):
     
     def inspect_constrained_nuisance_parameter(self, nuis, constraints):
         nuis_name = nuis.GetName()
-        _PRINT_.info('INFO: On nuisance parameter {}'.format(nuis_name))
+        self.stdout.info('INFO: On nuisance parameter {}'.format(nuis_name))
         nuip_nom = 0.0
         prefit_variation = 1.0
         found_constraint = ROOT.kFALSE
@@ -638,19 +678,19 @@ class ExtendedModel(object):
                             if math.isclose(old_sigma_value, 1.0, abs_tol=0.001):
                                 old_sigma_value = 1.0
                             if not found_sigma:
-                                _PRINT_.info('INFO: Sigma for pdf {} not found. Uisng 1.0.'.format(constr_name))
+                                self.stdout.info('INFO: Sigma for pdf {} not found. Uisng 1.0.'.format(constr_name))
                             else:
-                                _PRINT_.info('INFO: Uisng {} for sigma of pdf {}'.format(old_sigma_value, constr_name))
+                                self.stdout.info('INFO: Uisng {} for sigma of pdf {}'.format(old_sigma_value, constr_name))
 
                             prefit_variation = old_sigma_value
                         elif constraint.IsA() == ROOT.RooPoisson.Class():
                             constraint_type = 'pois'
                             tau = glob_obs.getVal()
-                            _PRINT_.info('INFO: Found tau {} of pdf'.format(constr_name))
+                            self.stdout.info('INFO: Found tau {} of pdf'.format(constr_name))
                             prefit_variation = 1. / math.sqrt(tau)
-                            _PRINT_.info('INFO: Prefit variation is {}'.format(prefit_variation))
+                            self.stdout.info('INFO: Prefit variation is {}'.format(prefit_variation))
                             nuip_nom = 1.0
-                            _PRINT_.info("INFO: Assume that {} is nominal value of the nuisance parameter".format(nuip_nom))
+                            self.stdout.info("INFO: Assume that {} is nominal value of the nuisance parameter".format(nuip_nom))
         return prefit_variation, constraint_type, nuip_nom
         
     def set_initial_errors(self, source:Optional["ROOT.RooArgSet"]=None):
@@ -662,7 +702,7 @@ class ExtendedModel(object):
             nuis_name = nuis.GetName()
             prefit_variation, constraint_type, _ = self.inspect_constrained_nuisance_parameter(nuis, all_constraints)
             if constraint_type=='gaus':
-                _PRINT_.info('INFO: Changing error of {} from {} to {}'.format(nuis_name, nuis.getError(), prefit_variation))
+                self.stdout.info('INFO: Changing error of {} from {} to {}'.format(nuis_name, nuis.getError(), prefit_variation))
                 nuis.setError(prefit_variation)
                 nuis.removeRange()    
         return None
@@ -670,7 +710,7 @@ class ExtendedModel(object):
     def get_poi(self, poi_name:Optional[str]=None, strict:bool=False):
         if poi_name is None:
             poi = self.pois.first()
-            _PRINT_.info(f"INFO: POI name not specified. The first POI `{poi.GetName()}` is used by default.")
+            self.stdout.info(f"INFO: POI name not specified. The first POI `{poi.GetName()}` is used by default.")
         else:
             poi = self.workspace.var(poi_name)
         if not poi:
@@ -777,10 +817,10 @@ class ExtendedModel(object):
             norm = pdf.getVal(obs)*target_obs.getBinWidth(i)
             n_events = norm*expected_events
             if n_events <= 0:
-                _PRINT_.warning("WARNING: Detected bin with zero expected events ({})! Please check"
+                self.stdout.warning("WARNING: Detected bin with zero expected events ({})! Please check"
                       "your inputs. Obs = {}, bin = {}".format(n_events, target_obs.GetName(), i))
             elif (n_events > 0) and (n_events < 1e18):
-                _PRINT_.debug("pdf={}, obs={}, bin={}, val={}".format(
+                self.stdout.debug("pdf={}, obs={}, bin={}, val={}".format(
                     pdf.GetName(), target_obs.GetName(), i, n_events))
                 asimov_data.add(self.observables, n_events)
             else:
@@ -791,16 +831,23 @@ class ExtendedModel(object):
         return asimov_data
     
     def match_globs(self):
-        nuis_list, globs_list, pdf_list = self.pair_constraints()
+        nuis_list, globs_list, pdf_list = self.pair_constraints(base_component=True)
+        for np, glob in zip(nuis_list, globs_list):
+            glob.setVal(np.getVal())
+            self.stdout.debug("set glob {} to val {}".format(glob.GetName(), glob.getVal()))
+        """
         for nuis, globs, pdf in zip(nuis_list, globs_list, pdf_list):
             pdf_class = pdf.ClassName()
             if pdf_class in ["RooGaussian", "RooBifurGauss"]:
                 globs.setVal(nuis.getVal())
             elif pdf_class == "RooPoisson":
+                from pdb import set_trace
+                set_trace()                
                 globs.setVal(nuis.getVal()*globs.getVal())
             else:
                 raise RuntimeError(f"constraint term `{pdf.GetName()}` has unsupported type `{pdf_class}`")
-            _PRINT_.debug("set glob {} to val {}".format(globs.GetName(), globs.getVal()))
+            self.stdout.debug("set glob {} to val {}".format(globs.GetName(), globs.getVal()))
+        """
 
     def generate_asimov(self, poi_name:str, poi_val:Optional[float]=None, 
                         poi_profile:Optional[float]=None,
@@ -812,6 +859,7 @@ class ExtendedModel(object):
                         channel_asimov_name:Optional[str]=None,
                         dataset:Optional["ROOT.RooDataSet"]=None,
                         constraint_option:int=0,
+                        restore_states:int=0,
                         minimizer_options:Optional[Dict]=None, 
                         nll_options:Optional[Union[Dict, List]]=None,
                         snapshot_names:Optional[Dict]=None):
@@ -849,6 +897,10 @@ class ExtendedModel(object):
                     Case 0: All nuisance parameters are allowed to float;
                     Case 1: Constrained nuisance parameters are fixed to 0. Unconstrained nuisrance
                             parameters are allowed to float.
+                restore_states: int, default=0
+                    Restore variable states at the end of asimov data generation.
+                    Case 0: All variable states will be restored.
+                    Case 1: Only global observable states will be restored.
                 do_import: bool, default=True
                     Import the generated asimov data to the current workspace.
                 asimov_name: (Optional) str
@@ -898,10 +950,10 @@ class ExtendedModel(object):
         # take snapshot of initial states of all variables
         ws.saveSnapshot(nom_vars_name, self.workspace.allVars())
         if not ws.getSnapshot(nom_glob_name):
-            _PRINT_.info('INFO: Saving snapshot "{}"'.format(nom_glob_name))
+            self.stdout.info('INFO: Saving snapshot "{}"'.format(nom_glob_name))
             ws.saveSnapshot(nom_glob_name, all_globs)
         if not ws.getSnapshot(nom_nuis_name):
-            _PRINT_.info('INFO: Saving snapshot "{}"'.format(nom_nuis_name))
+            self.stdout.info('INFO: Saving snapshot "{}"'.format(nom_nuis_name))
             ws.saveSnapshot(nom_nuis_name, all_nuis)
         
         if do_fit:
@@ -941,18 +993,8 @@ class ExtendedModel(object):
         poi_val = poi.getVal()
         
         # match values of global observables to the corresponding NPs
-        
         if modify_globs:
             self.match_globs()
-        
-        
-        """
-        if modify_globs:
-            constrained_nuis, constrained_globs = self.pair_nuis_and_glob_obs()
-            for np, glob in zip(constrained_nuis, constrained_globs):
-                glob.setVal(np.getVal())
-                _PRINT_.debug("set glob {} to val {}".format(glob.GetName(), glob.getVal()))
-        """
 
         if do_fit:
             ws.saveSnapshot(con_glob_name.format(mu=pretty_value(poi_profile)), all_globs)
@@ -987,17 +1029,57 @@ class ExtendedModel(object):
                                           ROOT.RooFit.WeightVar(weight_var))
         if do_import:
             getattr(ws, "import")(asimov_data)
-            _PRINT_.info(f'INFO: Generated Asimov Dataset `{asimov_data_name}`')
+            self.stdout.info(f'INFO: Generated Asimov Dataset `{asimov_data_name}`')
             
         if asimov_snapshot is not None:
-            _PRINT_.info('INFO: Saving snapshot "{}"'.format(asimov_snapshot))
+            self.stdout.info('INFO: Saving snapshot "{}"'.format(asimov_snapshot))
             ws.saveSnapshot(asimov_snapshot, self.workspace.allVars())
-        
-        # load back a snapshot of all varaible's initial states
-        ws.loadSnapshot(nom_vars_name)
+    
+        if restore_states == 0:
+            # load back a snapshot of all varaible's initial states
+            ws.loadSnapshot(nom_vars_name)
+        elif restore_states == 1:
+            # load back a snapshot of the initial global observable states
+            ws.loadSnapshot(nom_glob_name)
+        else:
+            raise ValueError(f"unsupported restore state option `{restore_states}`")
         
         return asimov_data
     
+    def generate_toys(self, n_toys:int=1, seed:int=0, binned:bool=True, 
+                      do_import:bool=True, name="toyData_{index}_seed_{seed}"):
+        ws = self.workspace
+        # take snapshot of initial states of all variables
+        self.workspace.saveSnapshot("tmp", self.workspace.allVars())
+        
+        if binned:
+            if self.pdf.ClassName() == "RooSimultaneous":
+                index_cat = self.pdf.indexCat()
+                for cat in index_cat:
+                    pdf_i = self.pdf.getPdf(cat.first)
+                    pdf_i.setAttribute("GenerateToys::Binned")
+            else:
+                self.pdf.setAttribute("GenerateToys::Binned")
+        
+        toys = []
+        args = [self.observables, ROOT.RooFit.Extended(), ROOT.RooFit.AutoBinned(True)]
+        if binned:
+            args.append(ROOT.RooFit.GenBinned("GenerateToys::Binned"))
+        
+        for i in range(n_toys):
+            self.randomize_globs(self.pdf, self.global_observables, seed)
+            toy = self.pdf.generate(*args)
+            toy_name = name.format(seed=seed, index=i)
+            toy.SetName(toy_name)
+            if do_import:
+                getattr(ws, "import")(toy)
+                self.stdout.info(f'INFO: Generated toy dataset "{toy_name}"')
+            toys.append(toy)
+        
+        ws.loadSnapshot("tmp")
+        
+        return toys
+    
     def save(self, fname:str, recreate:bool=True):
         self.workspace.writeToFile(fname, recreate)
     
@@ -1008,8 +1090,8 @@ class ExtendedModel(object):
         df = pd.DataFrame(data)
         return df
     
-    @staticmethod
-    def load_ws(fname:str, ws_name:Optional[str]=None, mc_name:Optional[str]=None):
+    @semistaticmethod
+    def load_ws(self, fname:str, ws_name:Optional[str]=None, mc_name:Optional[str]=None):
         if not os.path.exists(fname):
             raise FileNotFoundError('workspace file {} does not exist'.format(fname))
         file = ROOT.TFile(fname)
@@ -1021,7 +1103,7 @@ class ExtendedModel(object):
             if not ws_names:
                 raise RuntimeError("No workspaces found in the root file: {}".format(fname))
             if len(ws_names) > 1:
-                _PRINT_.warning("WARNING: Found multiple workspace instances from the root file: {}. Available workspaces"
+                self.stdout.warning("WARNING: Found multiple workspace instances from the root file: {}. Available workspaces"
                       " are \"{}\". Will choose the first one by default".format(fname, ','.join(ws_names)))
             ws_name = ws_names[0]
         ws = file.Get(ws_name)
@@ -1033,7 +1115,7 @@ class ExtendedModel(object):
             if not mc_names:
                 raise RuntimeError("no ModelConfig object found in the workspace: {}".format(ws_name))
             if len(mc_names) > 1:
-                _PRINT_.warning("WARNING: Found multiple ModelConfig instances from the workspace: {}. "
+                self.stdout.warning("WARNING: Found multiple ModelConfig instances from the workspace: {}. "
                       "Available ModelConfigs are \"{}\". "
                       "Will choose the first one by default".format(ws_name, ','.join(mc_names)))
             mc_name = mc_names[0]     
@@ -1165,8 +1247,8 @@ class ExtendedModel(object):
             distributions[cat]['y'] = y
         return distributions
     
-    @staticmethod
-    def get_simul_pdf_distributions(pdf, observables):
+    @semistaticmethod
+    def get_simul_pdf_distributions(self, pdf, observables):
         if not isinstance(pdf, ROOT.RooSimultaneous):
             raise ValueError("input pdf is not a simultaneous pdf")
         distributions = {}
@@ -1188,7 +1270,7 @@ class ExtendedModel(object):
                 n_events = norm*expected_events
                 bin_value = target_obs.getVal()
                 if n_events <= 0:
-                    _PRINT_.warning("WARNING: Detected bin with zero expected events ({})! "
+                    self.stdout.warning("WARNING: Detected bin with zero expected events ({})! "
                                     "Please check your inputs. Obs = {}, bin = {}".format(
                                     n_events, target_obs.GetName(), i))
                 elif (n_events > 0) and (n_events < 1e18):
