@@ -177,6 +177,10 @@ class Likelihood(AnalysisObject):
             return None
         return self.minimizer.fit_result.minNll()
     
+    @property
+    def status(self):
+        return self.minimizer.status
+    
     def __init__(self, filename:str, poi_name:Optional[str]=None,
                  data_name:str='combData', 
                  config:Optional[Union[Dict, str]]=None):
@@ -190,7 +194,7 @@ class Likelihood(AnalysisObject):
         self.poi = self.model.get_poi(poi_name)
         
     def evaluate(self, poi_val:float, vmin:float=10., vmax:float=10., 
-                 unconditional=False, hesse:bool=False, minos:bool=True, 
+                 unconditional=False, hesse:bool=False, minos:bool=False, 
                  snapshot_name:Optional[str]="initialSnapshot"):
         
         if snapshot_name:
@@ -202,15 +206,15 @@ class Likelihood(AnalysisObject):
         
         if unconditional:
             self.poi.setConstant(0)
-            hesse = True
         else:
             self.poi.setVal(poi_val)
             self.poi.setConstant(1)
 
-        minimizer_options = { 
+        minimizer_options = {
             'hesse': hesse,
             'minos': minos
         }
+        
         if minos:
             minimizer_options['minos_set'] = ROOT.RooArgSet(self.poi)
 
