@@ -84,7 +84,7 @@ class NuisanceParameterPull(object):
                    num_cpu:int=1, offset:bool=True, optimize:int=2, eigen:bool=False,
                    fix_cache:bool=True, fix_multi:bool=True, max_calls:int=-1, 
                    max_iters:int=-1, batch_mode:bool=False, int_bin_precision:float=-1.,
-                   outdir:str='output', cache=True, logfile_path=None, **kwargs):
+                   outdir:str='pulls', cache=True, logfile_path=None, **kwargs):
         # create output directory if not exists
         if not os.path.exists(outdir):
             os.makedirs(outdir, exist_ok=True)        
