@@ -7,6 +7,8 @@ import datetime
 import functools
 from concurrent.futures  import ProcessPoolExecutor
 
+import numpy as np
+
 def timely_info(green_text, normal_text, adjust=40):
     print(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), '\033[92m[INFO]\033[0m', '\033[92m{}\033[0m'.format(green_text).rjust(40, ' '), normal_text)
 
@@ -97,4 +99,15 @@ def parse_config(source:Optional[Union[Dict, str]]=None):
     elif isinstance(source, dict):
         return source
     else:
-        raise ValueError("invalid config input")
\ No newline at end of file
+        raise ValueError("invalid config input")
+        
+        
+class NpEncoder(json.JSONEncoder):
+    def default(self, obj):
+        if isinstance(obj, np.integer):
+            return int(obj)
+        if isinstance(obj, np.floating):
+            return float(obj)
+        if isinstance(obj, np.ndarray):
+            return obj.tolist()
+        return super(NpEncoder, self).default(obj)        
\ No newline at end of file
